# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pydantic import Field

from bazis.core.utils.schemas import BazisSettings


class Settings(BazisSettings):
    BAZIS_STORAGE_BG: str = Field('', title='Background task file storage engine')
    BAZIS_TASK_FOLDER: str = Field(
        'bg', title='Directory in the application where background tasks are located'
    )
    BAZIS_TASK_HANDLERS_LOCAL: int = Field(5, title='Minimum local number of handlers')
    BAZIS_TASK_HANDLERS_GLOBAL: int = Field(50, title='Global number of handlers')
    BAZIS_TASK_HANDLER_PORT: int = Field(49100, title='Default handler port')
    BAZIS_TASK_HANDLER_IDLE: float = Field(
        0.5, title='Handler wait time before a new work cycle'
    )
    BAZIS_TASK_SCHEDULER_IDLE: float = Field(
        0.2, title='Scheduler wait time before a new work cycle'
    )
    BAZIS_TASK_SCHEDULER_HOST: str | None = Field(None, title='Explicitly specifies the scheduler host')
    BAZIS_TASK_SCHEDULER_PORT: int = Field(49001, title='Default scheduler port')
    BAZIS_TASK_LIFETIME: int = Field(
        7776000, title='Lifetime of manual background tasks', dynamic=True
    )
    BAZIS_TASK_CRON_LIFETIME: int = Field(
        21600, title='Lifetime of periodic background tasks', dynamic=True
    )
    BAZIS_TASK_CRON_CLEANUP: int = Field(
        3600, title='Interval for running cleanup of periodic background tasks', dynamic=True
    )



settings = Settings()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.11 on 2025-07-28 14:02

from django.db import migrations, models
from django.utils.translation import gettext_lazy as _



class Migration(migrations.Migration):

    dependencies = [
        ('bg', '0002_alter_task_file'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='phase',
            field=models.CharField(max_length=1024, verbose_name='Phase'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta

from django.conf import settings
from django.utils.timezone import now

from ..basic.base import BgBase
from ..models import Task


class BgTasksClean(BgBase):
    name = 'Background tasks cleanup'
    parallel = 1

    def handle(self):
        # get all manual tasks that have not been updated for longer than the specified period
        qs = Task.objects.filter(
            dt_updated__lt=(now() - timedelta(seconds=settings.BAZIS_TASK_LIFETIME)),
            task_cron__isnull=True,
        )
        self.next_phase('Manual tasks cleanup', expected=qs.count())
        qs._raw_delete(qs.db)

        # get all periodic tasks that have not been updated for longer than the specified period
        qs = Task.objects.filter(
            dt_updated__lt=(now() - timedelta(seconds=settings.BAZIS_TASK_CRON_LIFETIME)),
            task_cron__isnull=False,
        )
        self.next_phase('Periodic tasks cleanup', expected=qs.count())
        qs._raw_delete(qs.db)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
from zipfile import ZipFile

from openpyxl import load_workbook


class XlsxReader:

    def __init__(self, fp, file_params):
        self.fp = fp
        self.fields_read = file_params.get('fields', [])
        self.titles = file_params.get('titles', {})
        self.groups_check = file_params.get('groups_check', False)
        self.files = None
        self.tmp_folder = None
        self.wb = None
        self.count = 0
        self.total_count = file_params.get('total_count', 0)

        if not self.titles:
            self.count_title_rows = 0
        elif 'groups' in self.titles and isinstance(self.titles['groups'], list):
            self.count_title_rows = 2
        else:
            self.count_title_rows = 1

    def event(self, e_type, *args, **kwargs):
        pass

    def get_titles(self):
        if not self.count_title_rows:
            return {f_name: f_name for f_name in self.fields_read}
        elif self.count_title_rows == 2:
            groups = self.titles['groups']
            titles = self.titles['titles']
            items = {}
            for f_name in self.fields_read:
                # find the group for the current column
                group = ([group for group in groups if f_name in group['columns']] + [None])[0]
                if group:
                    items[f_name] = group['name_unit'] % titles[f_name]
                else:
                    items[f_name] = titles[f_name]
            return items
        else:
            return {f_name: self.titles[f_name] for f_name in self.fields_read}

    def get_count(self):
        return self.count

    def __enter__(self):
        # determine the file extension
        ext = os.path.splitext(os.path.split(self.fp.name)[-1])[-1]

        if ext == '.xlsx':
            self.files = [self.fp.name]
        elif ext == '.zip':
            # in archive mode, groups cannot be processed
            self.groups_check = False
            # create a temporary folder
            self.tmp_folder = tempfile.mkdtemp()
            # extract the archive there
            with ZipFile(self.fp) as zp:
                zp.extractall(self.tmp_folder)
            # get the list of paths
            self.files = [os.path.join(self.tmp_folder, fn) for fn in os.listdir(self.tmp_folder)]
            self.event('zip', self.files)
        else:
            raise Exception('Unsupported file extension for upload')
        return self

    def __getitem__(self, key):
        result = None
        if isinstance(key, slice):
            result = []
            start = key.start
            stop = key.stop
            for i, it in enumerate(iter(self)):
                if i < start:
                    continue
                if i >= stop:
                    break
                result.append(it)
        else:
            for i, it in enumerate(iter(self)):
                if i == key:
                    result = it
        return result

    def __len__(self):
        return self.total_count

    def __iter__(self):
        fields_read = self.fields_read

        for f_path in self.files:
            # load the file
            self.wb = load_workbook(f_path, read_only=not self.groups_check, data_only=True)
            ws = self.wb.active
            count = (ws.max_row - self.count_title_rows) if ws.max_row else 0
            self.count += count

            self.event('file_read', f_path, count)

            row_num = 0
            rows = ws.iter_rows(values_only=True)
            # header row
            if self.count_title_rows:
                row_num += self.count_title_rows
                [next(rows) for _ in range(self.count_title_rows)]

            # if tracking groups is enabled, request them
            groups = ws.row_dimensions if self.groups_check else {}

            for row in rows:
                # skip completely empty rows
                if all([not col for col in row]):
                    continue

                row_num += 1

                row = (list(row) + ([None] * len(fields_read)))[: len(fields_read)]
                # main values
                data = dict(
                    zip(
                        [field for i, field in enumerate(fields_read)],
                        [val for j, val in enumerate(row)], strict=False,
                    )
                )
                # everything beyond the header length is loaded into a special key
                if len(fields_read) < len(row):
                    data['_ext'] = row[len(fields_read) :]
                # row metadata
                data['_meta'] = {'row_num': row_num}
                # if there are groups, get the nesting level
                if row_num in groups:
                    data['_meta']['group_level'] = groups.get(row_num).outline_level
                yield data
                self.event('row')
            self.wb.close()
            self.wb = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.wb:
            self.wb.close()

        # if a temporary folder was created, delete it
        if self.tmp_folder:
            shutil.rmtree(self.tmp_folder)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import socket
import sys
import traceback
from random import randint
from time import sleep, time

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils.timezone import now

from aiohttp import web

from bazis.core.utils.functools import throttle
from bazis.core.utils.network import healthcheck_server


sys.stdout.reconfigure(line_buffering=True)
sys.stderr.reconfigure(line_buffering=True)


LOG = logging.getLogger(__name__)


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--host', default=None, type=str)
        parser.add_argument('--port', default=None, type=int)

    def handle(self, host=None, port=None, **kwargs):
        TaskHandler = apps.get_model('bg.TaskHandler') # noqa F806
        Task = apps.get_model('bg.Task') # noqa F806

        host = host or settings.BAZIS_TASK_SCHEDULER_HOST or socket.gethostname()
        port = port or settings.BAZIS_TASK_HANDLER_PORT
        pid = f'{host}-{os.getpid()}'
        time_start = time()

        async def get_pid(r):
            return web.Response(text=pid)

        try:
            healthcheck_server(host, port, pid=get_pid)
        except Exception as e:
            LOG.error(str(e))
            return

        @throttle(5)
        def stop_check():
            return not TaskHandler.objects.filter(pid=pid).exists()

        @throttle(randint(3600 * 2, 3600 * 4), first_run=False)
        def stop_loop():
            return True

        # when starting, the handler is registered in the handlers table
        handler = TaskHandler.objects.create(
            pid=f'{pid}',
            host=host,
            port=port,
        )

        while not stop_check():

            if stop_loop():
                LOG.info(
                    f'Shutting down handler by time {pid}. Worked: {time() - time_start} sec'
                )
                break

            with transaction.atomic():
                for task in Task.objects.select_for_update(skip_locked=True).filter(
                    state='starting'
                )[:1]:
                    task.handler = handler
                    task.state = 'running'
                    task.phase = 'started'
                    task.dt_start = now()
                    task.save()
                    break
                else:
                    # delay
                    sleep(settings.BAZIS_TASK_HANDLER_IDLE)
                    continue

            # TODO: set ENVS
            LOG.info(f'Starting task {task.id} ({task.name})...')

            try:
                task.cls.run(task, *task.args, **task.kwargs)
            except Exception:  # NOQA [B902]
                task.set_done('Startup error', traceback.format_exc())
                LOG.error(task.error)
            else:
                if task := Task.objects.filter(id=task.id).first():
                    if task.state != 'done':
                        err = 'Uncaught execution error'
                        task.set_done(err, error=err)
                        LOG.error(f'Task {task.id}: {err}')
                    else:
                        LOG.info(f'Task {task.id} ({task.name}) completed successfully.')



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.core.exceptions import FieldDoesNotExist

from ..basic.base_download import BgBaseDownload


class BgBaseDownloadModel(BgBaseDownload):
    model = None

    @classmethod
    def get_name(cls):
        return f'Export: {cls.model._meta.verbose_name}'

    def get_titles(self):
        fields = {}
        for k in self.fields_read:
            try:
                fields[k] = self.model._meta.get_field(k).verbose_name
            except FieldDoesNotExist:
                fields[k] = k
        return fields

    def __init__(self):
        self.qs = self.model.objects.all()
        self.file_name = self.model.get_resource_name()

    def get_queryset(self):
        return self.qs

    def get_count(self):
        return self.qs.count()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.translation import gettext_lazy as _


class Command(BaseCommand):
    help = _('Initialize cleanup cron task')

    def handle(self, *args, **options):
        TaskCron = apps.get_model('bg.TaskCron') # noqa F806

        TaskCron.objects.get_or_create(
            cls_path='bazis.contrib.bg.bg.tasks_clean.BgTasksClean',
            defaults=dict(
                name=_('Clean tasks'),
                period=settings.BAZIS_TASK_CRON_CLEANUP,
            ),
        )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps
from django.utils.safestring import mark_safe


def task_is_run(cls):
    return apps.get_model('bg.Task').objects.filter(cls=cls).exclude(state='done').exists()


def task_is_success(task):
    return (task.state == 'done') and (not task.error) and (not task.interrupt)


def progress(task):
    if task.performed is None:
        return '-'

    d_names = dict(task.cls.d_names, **{'': 'Execution'})

    # convert to dicts
    expecteds = task.expected if isinstance(task.expected, dict) else {'': task.expected}
    performeds = task.performed if isinstance(task.performed, dict) else {'': task.performed}

    # determine intersection by keys
    slugs = set(list(expecteds.keys()) + list(performeds.keys()))

    # calculate progress bar height
    height = 35 if len(d_names) == 1 else 18
    font_size = 11 if len(d_names) == 1 else 9

    html = '<div style="position: relative; margin: -10px;">'
    for slug in slugs:
        expected = expecteds.get(slug, None) or 0
        performed = performeds.get(slug, None) or 0

        # if there is %, build the tag
        if expected:
            percent = f"""
                <div style="
                    position: absolute;
                    width: {(1.0 * performed / expected) * 100}%;
                    background-color: rgba(0, 255, 0, 0.5);
                    height: {height}px;
                    top: 0;
                    left: 0;
                    z-index: 1;"
                ></div>
            """
            count = f'{performed} / {expected}'
        else:
            percent = ''
            count = performed

        html += f"""
            <div style="position: relative;">
                <div style="
                    font-size: {font_size}px;
                    line-height: {height}px;
                    text-align: center;
                    font-weight: bold;
                    position: relative;
                    z-index: 2;
                    padding: 0 7px;
                ">{d_names[slug]}: {count}</div>
                {percent}
            </div>
        """

    html += '</div>'
    return mark_safe(html)


def history(task):
    d_names = dict(task.cls.d_names, **{'': 'Execution'})

    html = '<div>'
    for hist in task.phases_history:
        html += '<div>'
        html += '<div>{phase} ({duration} sec)</div>'.format(
            phase=hist['phase'],
            duration=hist['duration'],
        )

        # convert to dicts
        expecteds = (
            hist['expected'] if isinstance(hist['expected'], dict) else {'': hist['expected']}
        )
        performeds = (
            hist['performed'] if isinstance(hist['performed'], dict) else {'': hist['performed']}
        )

        # determine intersection by keys
        slugs = set(list(expecteds.keys()) + list(performeds.keys()))

        for slug in slugs:
            expected = expecteds.get(slug, None) or 0
            performed = performeds.get(slug, None) or 0

            # if there is %, build the tag
            if expected:
                count = f'{performed} / {expected}'
            else:
                count = performed

            if count:
                html += '<div>{name}: {count}</div>'.format(
                    name=d_names.get(slug, 'Not defined'),
                    count=count,
                )

        html += '<hr/></div>'
    html += '</div>'
    return mark_safe(html)


def spoiler(text, color):
    return mark_safe(
        f"""
        <a style="cursor: pointer; color: {color};"
           onclick="this.nextSibling.style.display = 'block'; this.style.display = 'none';"
       >Show</a><div style="display: none;">
            <a style="cursor: pointer;"
                onclick="this.parentElement.style.display='none';
                         this.parentElement.previousSibling.style.display='block';"
            >Hide</a>
            <pre style="color: {color}; font-size: 11px;">{text}</pre>
        </div>
    """
    )


def button_action(name, value, text, confirm=None):
    return mark_safe(
        '<button class="grp-button" style="width: auto;" name="{}" value="{}" {}>{}</button>'.format(
            name,
            value,
            (f'onclick="return confirm(\'{confirm}\')"') if confirm else '',
            text,
        )
    )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.core.management import call_command

import pytest
from entity.bg.calc_data import CalcData

from tests import factories


@pytest.mark.django_db(transaction=True)
def test_bg():
    factories.ParentEntityFactory.create_batch(15, child_entities=True)

    task = CalcData.delay()
    assert task.state == 'waiting'

    call_command('bg_task', task_id=task.id)

    task.refresh_from_db()
    assert task.state == 'done'

    phases_history = task.phases_history
    assert len(phases_history) == 1

    phase = phases_history[0]
    assert phase['expected'] == phase['performed'] == 15

    assert all(
        parent.is_active for parent in factories.ParentEntityFactory._meta.model.objects.all()
    )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
from zipfile import ZipFile

from openpyxl import load_workbook
from openpyxl.utils.cell import get_column_letter

from .base import BgBase


class BgBaseLoad(BgBase):
    fields_read = []
    skip_titles = False
    progress_count = 1000
    groups_check = False
    with_meta = False
    tmp_folder = None
    tmp_file = None

    def handle(self):
        # get the file
        fp = self.task_file_get()

        # determine the file extension
        ext = os.path.splitext(os.path.split(fp.name)[-1])[-1]

        # create a temporary local file, since the file may be taken from S3
        self.tmp_file = tempfile.NamedTemporaryFile(suffix=ext)
        self.tmp_file.write(fp.read())
        self.tmp_file.seek(0)

        if not fp:
            raise Exception('File not provided!')

        if ext == '.xlsx':
            files = [self.tmp_file.name]
            self.log.info('Received xlsx file: %s', fp.name)
        elif ext == '.zip':
            # in archive mode, groups cannot be processed
            self.groups_check = False
            # create a temporary folder
            self.tmp_folder = tempfile.mkdtemp()
            # extract the archive there
            with ZipFile(self.tmp_file) as zp:
                zp.extractall(self.tmp_folder)
            # get the list of paths
            files = [os.path.join(self.tmp_folder, fn) for fn in os.listdir(self.tmp_folder)]
            self.log.info('Received zip archive: %s', ', '.join(files))
        else:
            raise Exception('Unsupported file extension for upload')

        def reader():
            for f_path in files:
                # load the file
                wb = load_workbook(f_path, read_only=not self.groups_check, data_only=True)
                ws = wb.active
                count = (ws.max_row - int(self.skip_titles)) if ws.max_row else 0
                fields_read = self.fields_read

                self.next_phase(f'Reading file {f_path}', count)

                row_num = 0
                rows = ws.iter_rows(values_only=True)
                # header row
                if self.skip_titles:
                    heads = (list(next(rows)) + ([None] * len(fields_read)))[: len(fields_read)]
                    row_num += 1
                else:
                    heads = [get_column_letter(i + 1) for i in range(len(fields_read))]

                # if group tracking is enabled, request them
                groups = ws.row_dimensions if self.groups_check else {}

                for row in rows:
                    # skip completely empty row
                    if all([not col for col in row]):
                        continue

                    row_num += 1

                    row = (list(row) + ([None] * len(fields_read)))[: len(fields_read)]
                    # main values
                    data = dict(
                        zip(
                            [field for i, field in enumerate(fields_read) if heads[i]],
                            [val for j, val in enumerate(row) if heads[j]], strict=False,
                        )
                    )
                    # everything beyond the header length is loaded into a special key
                    if len(fields_read) < len(row):
                        data['_ext'] = row[len(fields_read) :]
                    # row metadata
                    if self.with_meta:
                        data['_meta'] = {
                            'row_num': row_num,
                        }
                        # if there are groups, get nesting
                        if row_num in groups:
                            data['_meta']['group_level'] = groups.get(row_num).outline_level
                    yield data
                    self.progress(d_count=self.progress_count)
                wb.close()

        self.load(reader())

    def load(self, reader):
        return

    def finishing(self):
        if self.tmp_folder:
            shutil.rmtree(self.tmp_folder)
        if self.tmp_file:
            self.tmp_file.close()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import os
from datetime import date, datetime
from io import BytesIO
from tempfile import NamedTemporaryFile
from uuid import UUID
from zipfile import ZIP_DEFLATED, ZipFile

from django.db.models import Manager, Model
from django.utils.functional import Promise
from django.utils.timezone import now

from openpyxl import Workbook, styles
from openpyxl.cell import WriteOnlyCell
from openpyxl.styles import Alignment
from openpyxl.utils import get_column_letter

from ..basic.base import BgBase


WRITABLE_HEAD_FONT = styles.Font(bold=True)
READABLE_FONT = styles.Font(sz=9)


class WBook:
    row_num = 0
    time = None
    _counter = None

    @classmethod
    def reset(cls):
        cls._counter = 0
        cls.time = now().strftime('%Y-%m-%d_%H-%M-%S')

    def __init__(self, parent):
        WBook._counter += 1

        # file object
        self.fp = BytesIO()
        self.parent = parent
        self.counter = WBook._counter
        self.unions = set()
        self.fields_read = parent.get_fields_read()
        self.fields_marked = parent.get_fields_marked()

        if parent.write_only:
            self.wb = Workbook(write_only=True)
            self.ws = self.wb.create_sheet()
        else:
            self.wb = Workbook()
            self.ws = self.wb.active

        titles = parent.get_titles()
        groups = []

        # prepare the header matrix
        gen_row_title = lambda: [None] * len(self.fields_read)  # NOQA [E731]
        if 'groups' in titles and isinstance(titles['groups'], list):
            rows_title = [gen_row_title(), gen_row_title()]
            groups = titles['groups']
            titles = titles['titles']
        else:
            rows_title = [gen_row_title()]

        for i, f_name in enumerate(self.fields_read):
            letter = get_column_letter(i + 1)
            # if column sizes are set - use them
            self.ws.column_dimensions[letter].width = parent.fields_size.get(f_name, 20)
            self.ws.column_dimensions[letter].auto_size = True
            # write Russian headers, if they exist
            if titles:
                # field cell
                cell = self.title_cell_bild(
                    titles.get(f_name, f_name), f_name in self.fields_marked
                )

                if groups:
                    # find the group for the current column
                    group = ([group for group in groups if f_name in group['columns']] + [None])[0]
                    if group:
                        rows_title[0][i] = self.title_cell_bild(
                            group['name_union'], group.get('marked')
                        )
                        rows_title[1][i] = cell
                        self.unions.add(
                            '{}1:{}1'.format(
                                self.get_letter_for_field(group['columns'][0]),
                                self.get_letter_for_field(group['columns'][-1]),
                            )
                        )
                    else:
                        rows_title[0][i] = cell
                        self.unions.add(f'{letter}1:{letter}2')
                else:
                    rows_title[0][i] = cell

        for row in rows_title:
            self.append(row)

    def title_cell_bild(self, value, marked=False):
        # convert lazy object to string in value
        c = WriteOnlyCell(ws=self.ws, value=str(value))
        if marked:
            c.font = WRITABLE_HEAD_FONT
        else:
            c.font = READABLE_FONT
        c.alignment = Alignment(wrapText=True, horizontal='center', vertical='center')
        return c

    def get_letter_for_field(self, field):
        try:
            return get_column_letter(self.fields_read.index(field) + 1)
        except ValueError:
            return None

    def append(self, item):
        self.row_num += 1
        self.ws.append(item)

    def get_name(self):
        return f'{self.parent.file_name}.{self.time}.xlsx'

    def save(self):
        for union in self.unions:
            self.ws.merged_cells.ranges.append(union)
        self.parent.pre_save()
        self.wb.save(self.fp)


class Zip:

    def __init__(self):
        self.zp = None
        self.fp = None

    def write(self, wbook):
        if not self.zp:
            self.fp = NamedTemporaryFile(mode='w+', prefix='bg_download_zip.', delete=False)
            self.zp = ZipFile(self.fp, mode="w", compression=ZIP_DEFLATED, allowZip64=True)

        self.zp.writestr(f'{wbook.counter}.{wbook.get_name()}', wbook.fp.getvalue())

    @property
    def is_init(self):
        return bool(self.zp)

    def close(self):
        self.zp.close()

    def unlink(self):
        if os.path.exists(self.fp.name):
            os.remove(self.fp.name)


class BgBaseDownload(BgBase):
    write_only = True
    is_data_ext = False
    fields_read = []
    fields_marked = set()
    fields_size = {}
    file_name = ''
    limit_rows_split = 500000
    parallel = 100

    def get_titles(self):
        return {}

    def get_fields_read(self):
        return self.fields_read

    def get_fields_marked(self):
        return self.fields_marked

    def get_queryset(self):
        return []

    def get_count(self):
        raise Exception('A specific number of results is expected')

    def pre_save(self):
        pass

    def handle(self):
        count = self.get_count()

        # presets
        WBook.reset()

        self.next_phase('Export', count)

        # references to structures
        fields_read = self.get_fields_read()
        qs = self.get_queryset()

        # create a new workbook
        self.wbook = WBook(self)

        # archive object
        _zip = Zip()

        # total number of rows
        total_count = 0
        # write each row
        row_count = 0

        for _i, row in enumerate(qs):
            if row_count >= self.limit_rows_split:
                # save data
                self.wbook.save()
                # write to zip
                _zip.write(self.wbook)
                # counters
                row_count = 0
                # create a new workbook
                self.wbook = WBook(self)

            # get a dictionary of values from the queryset record
            items = {
                k: (row.get(k, None) if isinstance(row, dict) else getattr(row, k, None))
                for k in fields_read
            }
            # standard values
            data = {}
            # extended values (can only exist if the corresponding key is the last one)
            data_ext = []

            # normalize
            for k, v in items.items():
                if isinstance(v, (str, Model, UUID)):
                    data[k] = str(v)
                elif isinstance(v, bool):
                    data[k] = int(v)
                elif isinstance(v, (Manager, list)):
                    # field values
                    if isinstance(v, Manager):
                        v_nested = [str(v_ext) for v_ext in v.all()]
                    else:
                        v_nested = v

                    if self.is_data_ext and k == fields_read[-1]:
                        data[k] = ''
                        data_ext = v_nested
                    else:
                        data[k] = ', '.join([str(x) for x in v_nested])
                elif isinstance(v, datetime):
                    data[k] = v.strftime('%Y-%m-%d %H:%M:%S')
                elif isinstance(v, date):
                    data[k] = v.strftime('%d.%m.%Y')
                elif isinstance(v, memoryview):
                    data[k] = str(v.tobytes())
                else:
                    data[k] = str(v)

            # assemble the list of values
            data = [data[key] for key in fields_read]
            if data_ext:
                data = data[:-1] + data_ext

            self.wbook.append(data)
            row_count += 1
            total_count += 1

            self.progress(d_count=1000)

        self.wbook.save()

        # if the archive is initialized
        if _zip.is_init:
            # write to the archive
            _zip.write(self.wbook)
            _zip.close()
            # and write it to the task as well
            self.task_file_save(
                f'{self.file_name}.{WBook.time}.zip',
                _zip.fp,
                file_params={
                    'fields': fields_read,
                    'titles': {
                        x: str(y) if isinstance(y, Promise) else y
                        for x, y in self.get_titles().items()
                    },
                    'total_count': total_count,
                },
            )
            # delete the temporary archive
            _zip.unlink()
        else:
            self.task_file_save(
                self.wbook.get_name(),
                self.wbook.fp,
                file_params={
                    'fields': fields_read,
                    'titles': {
                        x: str(y) if isinstance(y, Promise) else y
                        for x, y in self.get_titles().items()
                    },
                    'total_count': total_count,
                },
            )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.1 on 2024-08-23 23:26

import bazis.contrib.bg.models
import bazis.core.utils.orm
import django.core.files.storage.filesystem
import django.db.models.deletion
import model_clone.mixin
import uuid
from django.conf import settings
from django.db import migrations, models
from django.utils.translation import gettext_lazy as _



class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='TaskHandler',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pid', models.CharField(max_length=255, unique=True, verbose_name='PID')),
                ('host', models.CharField(max_length=255, verbose_name='Host')),
                ('port', models.PositiveIntegerField(verbose_name='Port')),
            ],
            options={
                'verbose_name': 'Handler',
                'verbose_name_plural': 'Handlers',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='TaskCron',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('cls_path', models.CharField(db_index=True, max_length=255, verbose_name='Task class')),
                ('args', models.JSONField(blank=True, null=True)),
                ('kwargs', models.JSONField(blank=True, null=True)),
                ('envs', models.JSONField(blank=True, null=True)),
                ('error', models.TextField(blank=True, null=True, verbose_name='Error')),
                ('period', models.CharField(max_length=100, verbose_name='Period value')),
                ('dt_run', models.DateTimeField(blank=True, null=True, verbose_name='Next run time')),
                ('is_enable', models.BooleanField(db_index=True, default=True, verbose_name='Enabled')),
                ('author', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_%(class)s', to=settings.AUTH_USER_MODEL)),
                ('author_updated', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_updated_%(class)s', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Periodic task',
                'verbose_name_plural': 'Periodic tasks',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('cls_path', models.CharField(db_index=True, max_length=255, verbose_name='Task class')),
                ('args', models.JSONField(blank=True, null=True)),
                ('kwargs', models.JSONField(blank=True, null=True)),
                ('envs', models.JSONField(blank=True, null=True)),
                ('error', models.TextField(blank=True, null=True, verbose_name='Error')),
                ('pid', models.CharField(db_index=True, max_length=36, verbose_name='PID')),
                ('state', models.CharField(choices=[('draft', 'Draft'), ('waiting', 'Waiting'), ('starting', 'Starting'), ('running', 'Running'), ('done', 'Completed')], db_index=True, default='draft', max_length=35, verbose_name='State')),
                ('phase', models.CharField(max_length=255, verbose_name='Phase')),
                ('phases_history', models.JSONField(blank=True, encoder=bazis.core.utils.orm.JsonFieldEncoder, null=True, verbose_name='History')),
                ('interrupt', models.BooleanField(db_index=True, default=False, verbose_name='Interrupted manually')),
                ('dt_start', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='Start time')),
                ('dt_finish', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='Finish time')),
                ('file', models.FileField(blank=True, null=True, storage=django.core.files.storage.filesystem.FileSystemStorage, upload_to=bazis.contrib.bg.models.task_file_name, verbose_name='File')),
                ('file_params', models.JSONField(blank=True, encoder=bazis.core.utils.orm.JsonFieldEncoder, null=True)),
                ('log', models.TextField(blank=True, null=True, verbose_name='Log')),
                ('expected', models.JSONField(blank=True, encoder=bazis.core.utils.orm.JsonFieldEncoder, null=True, verbose_name='Expected')),
                ('performed', models.JSONField(blank=True, encoder=bazis.core.utils.orm.JsonFieldEncoder, null=True, verbose_name='Completed')),
                ('result', models.JSONField(blank=True, encoder=bazis.core.utils.orm.JsonFieldEncoder, null=True, verbose_name='Result')),
                ('author', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_%(class)s', to=settings.AUTH_USER_MODEL)),
                ('author_updated', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='author_updated_%(class)s', to=settings.AUTH_USER_MODEL)),
                ('task_cron', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='bg.taskcron', verbose_name='Periodic task')),
                ('handler', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='bg.taskhandler')),
            ],
            options={
                'verbose_name': 'Background task',
                'verbose_name_plural': 'Background tasks',
                'ordering': ['-id'],
                'abstract': False,
                'index_together': {('cls_path', 'state'), ('state', 'dt_finish')},
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django import forms
from django.apps import apps
from django.contrib import admin, messages
from django.contrib.admin.helpers import ActionForm
from django.contrib.admin.options import IncorrectLookupParameters
from django.core.exceptions import ValidationError
from django.http import HttpResponseRedirect

# from .models import Task, TaskCron
from django.urls import reverse
from django.utils.safestring import mark_safe

from bazis.contrib.author.admin_abstract import AuthorAdminMixin
from bazis.core.admin_abstract import AutocompleteMixin, DtAdminMixin

from . import REGISTRY_BG_TASKS
from .admin_tags import button_action, history, progress, spoiler


class LoadActionForm(ActionForm):
    file_load = forms.FileField(required=False)


def create_task_action(task, func_task=None):
    def func(self, request, queryset):
        task_obj = func_task(task, request, queryset) if func_task else task.delay()

        # Dynamic URL generation
        admin_url = reverse('admin:bg_task_change', args=[task_obj.id])

        # print('task_obj:', task_obj)
        messages.add_message(
            request,
            messages.SUCCESS,
            f'Started: <a href="{admin_url}" target="_blank">{task.get_name()}</a>',
        )

    func.short_description = task.get_name()
    return func


def task_load_action(task, request, queryset):
    return task.delay(fp=request.FILES.get('file_load', None))


class LoadDownloadAdminMetaclass(forms.MediaDefiningClass):
    def __new__(cls, clsname, bases, dct):
        # file export actions
        for task in dct.get('tasks_download', []):
            action_name = f'action_{task.__name__.lower()}'
            dct[action_name] = create_task_action(task)
            dct.setdefault('actions_task_download', []).append(action_name)
            dct.setdefault('actions', []).append(action_name)

        # file import actions
        for task in dct.get('tasks_load', []):
            action_name = f'action_{task.__name__.lower()}'
            dct[action_name] = create_task_action(task, func_task=task_load_action)
            dct.setdefault('actions_task_load', []).append(action_name)
            dct.setdefault('actions', []).append(action_name)

        # task processing actions
        for task in dct.get('tasks_handler', []):
            action_name = f'action_{task.__name__.lower()}'
            dct[action_name] = create_task_action(task)
            dct.setdefault('actions_task_handler', []).append(action_name)
            dct.setdefault('actions', []).append(action_name)

        return super().__new__(cls, clsname, bases, dct)


class LoadDownloadAdminMixin(metaclass=LoadDownloadAdminMetaclass):
    action_form = LoadActionForm
    tasks_load = []
    tasks_download = []
    tasks_handler = []
    show_full_result_count = False

    def changelist_view(self, request, *args, **kwargs):
        action = request.POST.get('action', None)
        # background task actions
        actions = (
            getattr(self, 'actions_task_download', [])
            + getattr(self, 'actions_task_load', [])
            + getattr(self, 'actions_task_handler', [])
        )
        # if an action is passed, emulate data submission
        if action and action in actions and not request.POST.getlist('_selected_action'):
            request.POST = request.POST.copy()
            request.POST['_selected_action'] = 0
            request.POST['select_across'] = 1

        return super().changelist_view(request, *args, **kwargs)

    # hack to set list_editable if it is missing, for get_changelist_formset
    def get_changelist_instance(self, request):
        cl = super().get_changelist_instance(request)
        if not cl.list_editable:
            cl.list_editable = (None,)
        return cl

    # hack to set is_multipart
    def get_changelist_formset(self, request, **kwargs):
        formset = super().get_changelist_formset(request, **kwargs)
        formset.is_multipart = True
        return formset

    class Media:
        js = ('admin/js/jquery-3.5.1.min.js', 'admin/js/file_load.js')


class TaskDoneFilter(admin.SimpleListFilter):
    title = 'Completed tasks'
    parameter_name = 'done_state'

    def lookups(self, request, model_admin):
        return [
            ('success', 'Successful'),
            ('error', 'With error'),
            ('interrupt', 'Interrupted'),
        ]

    def queryset(self, request, qs):
        val = self.value()
        # search only among completed ones
        done = qs.filter(state='done')
        # check each case
        if val == 'success':
            return done.filter(error__isnull=True, interrupt=False)
        elif val == 'error':
            return done.filter(error__isnull=False)
        elif val == 'interrupt':
            return done.filter(interrupt=True)
        return qs


class TaskClsFilter(admin.SimpleListFilter):
    title = 'Types of background tasks'
    parameter_name = 'cls_path'

    def lookups(self, request, model_admin):
        items = []
        for bg_path, bg_cls in REGISTRY_BG_TASKS.items():
            items.append((bg_path, bg_cls.get_name()))
        return items

    def queryset(self, request, queryset):
        try:
            return queryset.filter(**self.used_parameters)
        except (ValueError, ValidationError) as e:
            # Fields may raise a ValueError or ValidationError when converting
            # the parameters to the correct type.
            raise IncorrectLookupParameters(e) from e


class TaskCronFilter(admin.RelatedFieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path, *args, **kwargs):
        # custom handling of an empty value
        without_cron = not params
        if without_cron:
            params['task_cron__isnull'] = ['True']
        super().__init__(field, request, params, model, model_admin, field_path)
        self.empty_value_display = 'Non-periodic'
        if without_cron:
            self.lookup_val_isnull = [True]
        # handling the special case of the entire selection
        if self.lookup_val == '0':
            self.used_parameters = {}

    def choices(self, changelist):
        for it in list(reversed(list(super().choices(changelist))))[:-1]:
            yield it
        it = {
            'selected': self.lookup_val == '0',
            'query_string': changelist.get_query_string(
                {self.lookup_kwarg: '0'},
                [self.lookup_kwarg_isnull],
            ),
            'display': 'All',
        }
        yield it


class TaskAdminBase(DtAdminMixin, AutocompleteMixin, AuthorAdminMixin, admin.ModelAdmin):
    list_per_page = 10
    list_display = (
        'name',
        'state',
        'phase',
        '_progress',
        '_phases_history',
        '_log',
        '_error',
        '_interrupting',
        '_file',
        'dt_start',
        'dt_finish',
    )
    readonly_fields = (
        'cls_path',
        'pid',
        'args',
        'kwargs',
        'state',
        'phase',
        'phases_history',
        'error',
        'interrupt',
        'dt_start',
        'dt_finish',
        'file',
        'log',
        'expected',
        'performed',
    )
    list_filter = (
        'state',
        TaskDoneFilter,
        TaskClsFilter,
        ('task_cron', TaskCronFilter),
    )

    def _file(self, task):
        if task.file:
            return mark_safe(f'<a href="{task.file.url}" download>{task.file.name}</a>')
        else:
            return '-'

    _file.short_description = 'Task file'

    def _error(self, task):
        return spoiler(task.error, 'red') if task.error else '-'

    _error.short_description = 'Error'

    def _interrupting(self, task):
        if not task.cls:
            return '-'

        if task.interrupt:
            return mark_safe('<div style="color: #fca400">Interrupted</div>')
        elif task.state == 'done':
            return '-'
        else:
            return button_action(
                'task_break',
                task.id,
                'Interrupt',
                confirm=f'ATTENTION! The task [{task.cls.get_name()}] will be interrupted!',
            )

    _interrupting.short_description = 'Interrupt'

    # progress of execution for the current phase
    def _progress(self, task):
        return progress(task)

    _progress.short_description = 'Progress'

    # progress of execution for the current phase
    def _phases_history(self, task):
        if not task.phases_history or not task.cls:
            return '-'
        return spoiler(history(task), 'blue')

    _phases_history.short_description = 'History'

    # execution log
    def _log(self, task):
        if not task.log:
            return '-'
        return spoiler(task.log, 'blue')

    _log.short_description = 'Log'

    def changelist_view(self, request, extra_context=None):
        if 'task_break' in request.POST:
            # get the task
            task = apps.get_model('bg.Task').objects.get(id=request.POST['task_break'])
            # interrupt execution
            task.interrupt = True
            # if the task is waiting to start - immediately change the status
            if task.state in ('waiting', 'starting'):
                task.state = 'done'
                task.phase = 'interrupted'
            # save the status
            task.save(update_fields=['interrupt', 'state', 'phase'])
            return HttpResponseRedirect(request.get_full_path())
        return super().changelist_view(request, extra_context)


class TaskCronForm(forms.ModelForm):
    cls_path = forms.ChoiceField(label='Background task')

    class Meta:
        model = apps.get_model('bg.TaskCron')
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['cls_path'].choices = [
            (bg_path, bg_cls.get_name()) for bg_path, bg_cls in REGISTRY_BG_TASKS.items()
        ]


class TaskCronAdminBase(DtAdminMixin, AutocompleteMixin, AuthorAdminMixin, admin.ModelAdmin):
    list_display = ('name', 'cls_path', 'args', 'kwargs', 'period', 'dt_run', 'is_enable')
    list_editable = ('is_enable',)
    readonly_fields = ('dt_run',)
    form = TaskCronForm



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps
from django.core.management.base import BaseCommand


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--task_id', action='store', dest='task_id', default=None)

    def handle(self, task_id, **kwargs):
        Task = apps.get_model('bg.Task') # noqa F806
        task = Task.objects.get(id=task_id)
        task.cls.run(task, *task.args, **task.kwargs)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.db import models
from django.utils.translation import gettext_lazy as _

from bazis_test_utils.models_abstract import (
    ChildEntityBase,
    DependentEntityBase,
    ExtendedEntityBase,
    ParentEntityBase,
)

from bazis.core.models_abstract import DtMixin, JsonApiMixin, UuidMixin


class ChildEntity(DtMixin, UuidMixin, JsonApiMixin, ChildEntityBase):
    class Meta:
        verbose_name = _('Child entity')
        verbose_name_plural = _('Child entities')


class DependentEntity(DtMixin, UuidMixin, JsonApiMixin, DependentEntityBase):
    parent_entity = models.ForeignKey(
        'ParentEntity', on_delete=models.CASCADE, related_name='dependent_entities'
    )

    class Meta:
        verbose_name = _('Dependent entity')
        verbose_name_plural = _('Dependent entities')


class ExtendedEntity(DtMixin, UuidMixin, JsonApiMixin, ExtendedEntityBase):
    parent_entity = models.OneToOneField(
        'ParentEntity', on_delete=models.CASCADE, related_name='extended_entity'
    )

    class Meta:
        verbose_name = _('Extended entity')
        verbose_name_plural = _('Extended entities')


class ParentEntity(DtMixin, UuidMixin, JsonApiMixin, ParentEntityBase):
    s = models.TextField
    child_entities = models.ManyToManyField(
        ChildEntity,
        related_name='parent_entities',
        blank=True,
    )

    class Meta:
        verbose_name = _('Parent entity')
        verbose_name_plural = _('Parent entities')



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.db import models
from django.utils.functional import cached_property
from django.utils.timezone import now

from bazis.contrib.author.models_abstract import AuthorMixin
from bazis.core.models_abstract import DtMixin, InitialBase, UuidMixin
from bazis.core.utils.imp import import_class
from bazis.core.utils.orm import JsonFieldEncoder


def task_file_name(obj, fn):
    return f'{settings.BAZIS_TASK_FOLDER}/{fn}'


if settings.BAZIS_STORAGE_BG:
    TaskStorage = import_class(settings.BAZIS_STORAGE_BG)
else:
    TaskStorage = FileSystemStorage


class TaskBaseModel(UuidMixin, DtMixin, AuthorMixin):
    name = models.CharField('Name', max_length=255)
    cls_path = models.CharField('Task class', max_length=255, db_index=True)
    args = models.JSONField(null=True, blank=True)
    kwargs = models.JSONField(null=True, blank=True)
    envs = models.JSONField(null=True, blank=True)
    error = models.TextField('Error', null=True, blank=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    @cached_property
    def cls(self):
        try:
            return import_class(self.cls_path)
        except Exception:
            return


class TaskBase(TaskBaseModel):
    STATES = [
        ('draft', 'Draft'),
        ('waiting', 'Waiting'),
        ('starting', 'Starting'),
        ('running', 'Running'),
        ('done', 'Completed'),
    ]
    pid = models.CharField('PID', max_length=36, db_index=True)
    handler = models.ForeignKey('TaskHandler', blank=True, null=True, on_delete=models.SET_NULL)
    state = models.CharField(
        'State', max_length=35, choices=STATES, default=STATES[0][0], db_index=True
    )
    phase = models.CharField('Phase', max_length=1024)
    phases_history = models.JSONField('History', encoder=JsonFieldEncoder, null=True, blank=True)
    interrupt = models.BooleanField('Interrupted manually', default=False, db_index=True)
    dt_start = models.DateTimeField('Start time', null=True, blank=True, db_index=True)
    dt_finish = models.DateTimeField('Finish time', null=True, blank=True, db_index=True)
    file = models.FileField(
        'File', upload_to=task_file_name, blank=True, null=True
    )
    file_params = models.JSONField(null=True, encoder=JsonFieldEncoder, blank=True)
    log = models.TextField('Log', blank=True, null=True)
    expected = models.JSONField('Expected', encoder=JsonFieldEncoder, null=True, blank=True)
    performed = models.JSONField('Performed', encoder=JsonFieldEncoder, null=True, blank=True)
    task_cron = models.ForeignKey(
        'TaskCron',
        verbose_name='Periodic task',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    result = models.JSONField('Result', encoder=JsonFieldEncoder, null=True, blank=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._meta.get_field('file').storage = TaskStorage()

    class Meta:
        abstract = True
        verbose_name = 'Background task'
        verbose_name_plural = 'Background tasks'
        ordering = ['-id']
        index_together = [
            ('cls_path', 'state'),
            ('state', 'dt_finish'),
        ]

    def set_done(self, phase, error=None):
        self.dt_finish = now()
        self.error = error
        self.phase = phase
        self.state = 'done'
        self.save(update_fields=['state', 'phase', 'error', 'dt_finish'])

    @property
    def is_done(self):
        return self.state == 'done'

    @property
    def is_success(self):
        return (self.state == 'done') and (not self.error) and (not self.interrupt)

    @property
    def is_error(self):
        return (self.state == 'done') and self.error

    @property
    def is_interrupt(self):
        return (self.state == 'done') and self.interrupt


class TaskCronBase(TaskBaseModel):
    period = models.CharField('Period value', max_length=100)
    dt_run = models.DateTimeField('Next run time', null=True, blank=True)
    is_enable = models.BooleanField('Enabled', default=True, db_index=True)

    class Meta:
        abstract = True
        verbose_name = 'Periodic task'
        verbose_name_plural = 'Periodic tasks'


class TaskHandlerBase(UuidMixin, InitialBase):
    class Meta:
        abstract = True
        verbose_name = 'Handler'
        verbose_name_plural = 'Handlers'

    pid = models.CharField('PID', max_length=255, unique=True)
    host = models.CharField('Host', max_length=255)
    port = models.PositiveIntegerField('Port')



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from io import BytesIO, StringIO

from django.core.management import call_command
from django.db.models import Model
from django.utils.timezone import now

from bazis.contrib.bg.basic.base import BgBase


class FixtureDownload(BgBase):
    name = 'Fixture export'
    parallel = 1
    models_export: list[type[Model] | str] = []
    file_name = 'fixtures'

    def handle(self):
        models_export = [
            m if isinstance(m, str) else f'{m._meta.app_label}.{m._meta.model_name}'
            for m in self.models_export
        ]

        target_models = ", ".join(models_export)
        self.next_phase(f'Downloading fixtures for models: {target_models}'[:512])

        fp = StringIO()
        call_command(
            'dumpdata',
            *models_export,
            natural_foreign=True,
            natural_primary=True,
            stdout=fp,
        )
        fp.seek(0)

        data = json.load(fp)
        fp = BytesIO()
        fp.write(json.dumps(data, ensure_ascii=False).encode())

        self.task_file_save(f"{now().strftime('%Y-%m-%d_%H-%M-%S')}.{self.file_name}.json", fp)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps

from bazis.core.apps import BaseConfig
from bazis.core.utils.imp import import_module, pkg_load


class BgConfig(BaseConfig):
    name = 'bazis.contrib.bg'
    verbose_name = 'Background tasks'

    def ready(self):
        super().ready()

        # perform loading of bg packages of connected applications
        for app in apps.get_app_configs():
            try:
                bg_pkg = import_module(f'{app.module.__name__}.bg')
            except ModuleNotFoundError:
                pass
            else:
                pkg_load(bg_pkg)




# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps

from bazis.core.routes_abstract.jsonapi import JsonapiRouteBase
from bazis.core.schemas import SchemaFields


class ChildEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ChildEntity')

    fields = {
        None: SchemaFields(
            include={
                'parent_entities': None,
            },
        ),
    }


class DependentEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.DependentEntity')


class ExtendedEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ExtendedEntity')


class ParentEntityRouteSet(JsonapiRouteBase):
    model = apps.get_model('entity.ParentEntity')

    # add fields (extended_entity, dependent_entities) to schema
    fields = {
        None: SchemaFields(
            include={'extended_entity': None, 'dependent_entities': None},
        ),
    }



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback
from io import BufferedIOBase, StringIO
from time import sleep, time

from django.apps import apps
from django.db import transaction
from django.utils.timezone import now

from .. import REGISTRY_BG_TASKS


LOG = logging.getLogger(__name__)


class InterruptException(Exception):  # NOQA [N818]
    pass


class BgLogHandler(logging.StreamHandler):

    def __init__(self, io):
        super().__init__(io)
        # set formatter
        self.formatter = logging.Formatter("%(levelname)s:%(asctime)s:%(message)s")


def commit(using=None):
    if not transaction.get_connection(using).in_atomic_block:
        transaction.commit(using)


def rollback(using=None):
    if not transaction.get_connection(using).in_atomic_block:
        transaction.rollback(using)


def set_autocommit(autocommit, using=None):
    if not transaction.get_connection(using).in_atomic_block:
        transaction.set_autocommit(autocommit, using=None)


class BgBaseMeta(type):

    def __new__(mcs, name: str, bases: tuple, attrs: dict, **kwargs):
        # create class
        bg_cls: type[BgBase] = super().__new__(mcs, name, bases, attrs, **kwargs)

        # determine task name
        try:
            bg_name = bg_cls.get_name()
        except Exception:
            bg_name = None
        # further logic only for concrete classes
        if not bg_name:
            return bg_cls

        REGISTRY_BG_TASKS[bg_cls.get_path()] = bg_cls

        return bg_cls


class BgBase(metaclass=BgBaseMeta):
    # human-readable task name
    name = None
    # number of simultaneously running tasks
    parallel = 1
    # tasks that block execution
    blocked = []
    # naming dictionary
    d_names = {
        '': 'Execution',
    }
    # whether it should be processed in a transaction
    is_transaction = True
    # current task id
    task_id = None

    log = None

    @classmethod
    def get_name(cls):
        return cls.name

    @classmethod
    def get_path(cls):
        return f'{cls.__module__}.{cls.__qualname__}'

    @classmethod
    def delay(cls, *args, **kwargs):
        """
        This method is called in the application context
        """
        # try to extract the name
        name = kwargs.pop('name', None)
        # try to extract the passed file
        fp = kwargs.pop('fp', None)
        # try to extract the environment dictionary
        envs = kwargs.pop('envs', None)
        # try to get the task author
        author = kwargs.pop('author', None)

        # create a task with a unique id
        task = apps.get_model('bg.Task').objects.create(
            name=name or cls.get_name(),
            cls_path=cls.get_path(),
            phase='draft',
            phases_history=[],
            args=args,
            kwargs=kwargs,
            envs=envs,
            author=author,
        )
        if fp:
            task.file.save(fp.name, fp)

        # start the task
        task.state = 'waiting'
        task.phase = 'waiting for start'
        task.save()

        return task

    @classmethod
    def delay_sync(cls, *args, **kwargs):
        """
        Starts and waits for completion
        """
        task = cls.delay(*args, **kwargs)
        # wait for completion
        while not apps.get_model('bg.Task').objects.get(id=task.id).state == 'done':
            sleep(1)
        return task

    @classmethod
    def run(cls, task, *args, **kwargs): # noqa [C901]
        """
        This method is called in the context of a separate process
        """
        error = None
        obj = None

        try:
            obj = cls(*args, **kwargs)

            obj.task_id = task.id
            obj.task_name = task.name
            # start of the phase
            obj._start = None
            # phase name
            obj._phase_name = None
            # timestamp of the last update
            obj._last_sync = None
            # counter
            obj._i = None
            # actual expected progress
            obj._expected = None
            # actual performed progress
            obj._performed = None
            # create a logger specifically for the given task
            obj._log = StringIO()
            obj.log = logging.getLogger()
            obj.log.setLevel(logging.INFO)
            obj.log.addHandler(BgLogHandler(obj._log))

            # disable autocommit
            if cls.is_transaction:
                set_autocommit(False)

            # run the main class logic
            try:
                obj.pre_handle()
                obj.handle()
                obj.post_handle()
            except Exception:  # NOQA [B902]
                # roll back
                if cls.is_transaction:
                    rollback()
                raise
            else:
                if cls.is_transaction:
                    commit()
            finally:
                # enable autocommit
                if cls.is_transaction:
                    set_autocommit(True)
        except InterruptException:
            phase = 'interrupted'
            # allow the child class to properly handle interruption
            if obj:
                try:
                    obj.interrupting()
                except Exception:  # NOQA [B902]
                    LOG.fatal(
                        'Interrupting error in class (%s, %s): %s',
                        cls.__name__,
                        task.id,
                        traceback.format_exc(),
                    )
        except Exception:  # NOQA [B902]
            phase = 'error'
            error = traceback.format_exc()
            LOG.fatal('Exception in class (%s, %s): %s', cls.__name__, task.id, error)
            # allow the child class to properly handle the exception
            if obj:
                try:
                    obj.excepting()
                except Exception:  # NOQA [B902]
                    LOG.fatal(
                        'Excepting error in class (%s, %s): %s',
                        cls.__name__,
                        task.id,
                        traceback.format_exc(),
                    )
        else:
            phase = 'completed'
        finally:
            if obj:
                # save the last state of the phase
                try:
                    obj._phase_save()
                except Exception:  # NOQA [B902]
                    LOG.error(
                        'Finish _phase_save error in class (%s, %s): %s',
                        cls.__name__,
                        task.id,
                        traceback.format_exc(),
                    )

                # allow child class to correctly handle unconditional completion
                try:
                    obj.finishing()
                except Exception:  # NOQA [B902]
                    LOG.fatal(
                        'Finishing error in class (%s, %s): %s',
                        cls.__name__,
                        task.id,
                        traceback.format_exc(),
                    )

        # get the task for completion
        task = apps.get_model('bg.Task').objects.get(id=task.id)
        task.set_done(phase, error)

    def interrupting(self):
        pass

    def excepting(self):
        pass

    def finishing(self):
        pass

    def set_result(self, result):
        self.task_update(result=result)

    def get_task(self):
        return apps.get_model('bg.Task').objects.get(id=self.task_id)

    def task_update(self, **kwargs):
        # commit to synchronize the DB
        if self.is_transaction:
            commit()

        # simple update
        task = self.get_task()
        for k, v in kwargs.items():
            setattr(task, k, v)

        # perform a real update only when there is data
        if kwargs:
            task.save(update_fields=kwargs.keys())

        # check whether the task has been interrupted
        if task.interrupt:
            self.log.warning('The task was interrupted!')
            raise InterruptException

        # commit
        if self.is_transaction:
            commit()

        # return the task object
        return task

    def task_file_save(self, name: str, fp: BufferedIOBase, file_params: dict = None):
        """
        Saves a file in the task object.

        :param name: File name
        :param fp: file descriptor
        :param file_params: dictionary of additional file data
        """
        # save the file
        task = self.get_task()
        task.file.save(name, fp)
        if file_params:
            return self.task_update(file_params=file_params)
        return task

    def task_file_get(self) -> BufferedIOBase | None:
        """
        Returns the saved file descriptor from the task
        """
        task = self.get_task()
        return task.file.file if task.file else None

    def next_phase(self, phase_name: str, expected: int | dict[str, int] = None):
        """
        Sets a new task phase

        :param phase_name: String name of the phase
        :param expected: expected number of operations. Can be a dictionary
        :return:
        """
        # save the last status state
        self._phase_save()

        self.task_update(phase=phase_name, expected=expected)
        self.log.info('%s. Expected count: %s', phase_name, expected or 'not defined')

        # initialize phase variables
        self._phase_name = phase_name
        self._start = now()
        self._last_sync = time()
        self._i = 0
        self._expected = expected
        self._performed = None

    def flush(self):
        """
        Synchronizes data with the DB
        """
        # update progress and optionally change the expected number of operations
        self.task_update(
            performed=self._performed,
            expected=self._expected,
            log=self._log.getvalue(),
        )
        # update the last synchronization time
        self._last_sync = time()

    def progress(
        self,
        performed: int | dict[str, int] = None,
        expected: int | dict[str, int] = None,
        d_time: int = None,
        d_count: int = None,
        need_sync: bool = False,
    ):
        """
        Sets the execution and expectation counters.
        Checks and synchronizes data with the DB

        :param performed: current number of completed operations. Can be a dictionary
        :param expected: expected number of operations. Can be a dictionary
        :param d_time: synchronization period, seconds
        :param d_count: if set, also limits the synchronization period by the number of operations
        :param need_sync: if True, forces synchronization with the DB
        :return:
        """
        # increase counter
        self._i += 1
        # save actual progress
        self._performed = performed if performed is not None else self._i
        # save actual expectation
        if expected:
            self._expected = expected

        # 3 seconds - default synchronization period
        d_time = d_time or 3

        # always decide based on d_time. If d_count is set, it also limits the time checks
        if not need_sync and (not d_count or (self._i % d_count) == 0):
            need_sync = (time() - self._last_sync) > d_time

        if need_sync:
            self.flush()

    def pre_handle(self):
        pass

    def handle(self):
        raise Exception('handle no implement!')

    def post_handle(self):
        pass

    def _phase_save(self):
        """
        Save the status to history along with its execution time
        """
        task = self.get_task()
        task.log = self._log.getvalue()

        if self._start:
            task.performed = None
            task.expected = None
            task.phases_history.append(
                {
                    'phase': self._phase_name,
                    'duration': round((now() - self._start).total_seconds(), 1),
                    'expected': self._expected,
                    'performed': self._performed,
                }
            )
            self.log.info('Phase recorded: %s', self._phase_name)

        task.save(update_fields=['performed', 'expected', 'phases_history', 'log'])
        # commit
        if self.is_transaction:
            commit()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.utils.translation import gettext_lazy as _

from bazis.contrib.bg.basic.base import BgBase

from entity.models import ParentEntity


class CalcData(BgBase):
    name = _('Calc data')
    parallel = 1

    def handle(self) -> None:
        qs = ParentEntity.objects.all()
        self.next_phase(phase_name='First phase: parents', expected=qs.count())

        for parent in ParentEntity.objects.all():
            parent.is_active = True
            parent.save()
            self.progress()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import socket
import sys
import threading
import traceback
from collections import defaultdict
from datetime import datetime, timedelta
from time import sleep

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Exists, OuterRef, Q, Subquery
from django.utils.timezone import make_aware, now

import aiohttp
import psutil
from asgiref.sync import sync_to_async
from crontab import CronTab

from bazis.core.utils.functools import throttle
from bazis.core.utils.network import check_port, healthcheck_server


sys.stdout.reconfigure(line_buffering=True)
sys.stderr.reconfigure(line_buffering=True)

LOG = logging.getLogger(__name__)


def proc_start(cmd, envs=None):
    env = os.environ.copy()
    env.update(envs or {})
    return psutil.Popen([str(x) for x in cmd], env=env)


class Command(BaseCommand):

    @staticmethod
    def _task_cron_update(task_cron, **kwargs):
        for k, v in kwargs.items():
            setattr(task_cron, k, v)
        task_cron.save(update_fields=kwargs.keys())

    @throttle(5)
    def cron_check(self):
        TaskCron = apps.get_model('bg.TaskCron') # noqa F806
        Task = apps.get_model('bg.Task') # noqa F806

        # iterate over all cron tasks
        for task_cron in TaskCron.objects.filter(is_enable=True).annotate(
            is_task_executing=Exists(
                Task.objects.filter(
                    Q(
                        task_cron=OuterRef('pk'),
                    )
                    & ~Q(state='done')
                ).values('id')
            ),
            task_dt_finish=Subquery(
                Task.objects.filter(
                    Q(
                        task_cron=OuterRef('pk'),
                    )
                    & Q(state='done')
                )
                .order_by('-dt_finish')
                .values('dt_finish')[:1]
            ),
        ):
            task_cls = task_cron.cls
            if not task_cls:
                print(f'Cron task class not found: {task_cron.cls_path}')
                continue

            # if the next run time is specified - check it
            if task_cron.dt_run:
                if task_cron.dt_run <= now():
                    # create a specific task
                    kwargs = dict(
                        task_cron.kwargs or {},
                        **{
                            'envs': task_cron.envs,
                        },
                    )
                    # actually start the task, preserving the original cron task
                    task = task_cls.delay(*(task_cron.args or []), **kwargs)
                    task.task_cron = task_cron
                    task.save(update_fields=['task_cron'])

                    # reset the next run time
                    self._task_cron_update(task_cron, dt_run=None)
                # in any case - no more checks are needed in this iteration
                continue

            # if it is assumed that the run will be periodic
            if task_cron.period.isnumeric():
                # then the period is specified in seconds
                seconds = float(task_cron.period)
                # if the task is currently running - no new run is required
                if task_cron.is_task_executing:
                    continue
                # get the last completed task
                # set the next run time, counting from the completion time of the previous task
                self._task_cron_update(
                    task_cron,
                    dt_run=(task_cron.task_dt_finish or now()) + timedelta(seconds=seconds)
                )
            else:
                # try to create a cron object
                try:
                    cron = CronTab(task_cron.period)
                except Exception:  # NOQA [B902]
                    # in case of an error, write it to the task
                    self._task_cron_update(task_cron, error=traceback.format_exc())
                else:
                    # write the time of the next run to the task
                    self._task_cron_update(
                        task_cron,
                        dt_run=make_aware(
                            datetime.fromtimestamp(cron.next(default_utc=True, delta=False)),
                        )
                    )

    @throttle(2)
    def tasks_check(self):
        Task = apps.get_model('bg.Task') # noqa F806

        # explicitly disable all tasks in the DB that are linked to handlers missing from the table
        Task.objects.filter(
            Q(state='running') | Q(state='done', dt_finish=None), Q(handler=None)
        ).update(
            phase='Unhandled interruption',
            error='The task was interrupted for an unknown reason',
            state='done',
            dt_finish=now(),
        )

    def tasks_prepare(self):
        Task = apps.get_model('bg.Task') # noqa F806

        # collect running tasks according to the DB
        actives_cls = defaultdict(set)
        for t in Task.objects.filter(state__in=['running', 'starting'])[:10]:
            actives_cls[t.cls_path].add(t)

        # iterate over all tasks waiting to be started
        with transaction.atomic():
            for task in Task.objects.filter(state='waiting')[:10]:
                if not task.cls:
                    task.delete()
                    continue

                # if the task has already been interrupted - close it
                if task.interrupt:
                    task.set_done('interrupted')
                    continue

                # rejection status
                phase_reject = None

                # identity check
                if any(
                    [
                        t.args == task.args and t.kwargs == task.kwargs
                        for t in actives_cls[task.cls_path]
                    ]
                ):
                    phase_reject = 'The task is already running. Waiting for completion'
                # if the number of active tasks has exceeded the allowed level
                elif len(actives_cls[task.cls_path]) >= task.cls.parallel:
                    phase_reject = f'Maximum of {task.cls.parallel} similar tasks'
                else:
                    # determine blocking tasks
                    blocked = [
                        cls_block
                        for cls_block in task.cls.blocked
                        if len(actives_cls[cls_block]) > 0
                    ]
                    if blocked:
                        phase_reject = 'Waiting for completion of tasks: {}'.format(', '.join(
                            [cls_block for cls_block in blocked]
                        ))

                # if there is a rejection status, and the current task status is different - update the status
                if phase_reject:
                    # update the status only if it has been changed
                    if task.phase != phase_reject:
                        task.phase = phase_reject
                        task.save(update_fields=['phase'])
                    continue

                # mark the task for launch
                task.state = 'starting'
                task.save(update_fields=['state'])

                actives_cls[task.cls_path].add(task)

    def start_handlers_check(self):
        TaskHandler = apps.get_model('bg.TaskHandler') # noqa F806

        async def main_thread_loop_work():
            while True:
                handlers = [handler async for handler in TaskHandler.objects.all()]

                async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=5)) as session:
                    items = [
                        session.get(f'http://{handler.host}:{handler.port}/pid')
                        for handler in handlers
                    ]
                    results = await asyncio.gather(*items, return_exceptions=True)

                for_delete = [
                    handlers[i].pk
                    for i, r in enumerate(results)
                    if isinstance(r, Exception)
                    or r.status != 200
                    or (await r.text()) != handlers[i].pid
                ]

                await sync_to_async(TaskHandler.objects.filter(pk__in=for_delete).delete)()

                await asyncio.sleep(3)

        def thead_worker():
            asyncio.run(main_thread_loop_work())

        # start a separate thread to poll handlers
        th = threading.Thread(target=thead_worker, daemon=True)
        th.start()

    @throttle(30)
    def run_handlers(self):
        TaskHandler = apps.get_model('bg.TaskHandler') # noqa F806

        ports_inuse = set(TaskHandler.objects.values_list('port', flat=True))

        for _ in range(settings.BAZIS_TASK_HANDLERS_LOCAL - len(ports_inuse)):
            for i in range(1000):
                port = settings.BAZIS_TASK_HANDLER_PORT + i
                if port not in ports_inuse:
                    if check_port(port):
                        LOG.info(f'Handler localhost:{port} has been ran')
                        proc_start(
                            [
                                sys.executable,
                                'manage.py',
                                'bg_handler',
                                '--host',
                                '127.0.0.1',
                                '--port',
                                port,
                            ]
                        )
                        ports_inuse.add(port)
                        break

    def handle(self, **kwargs):
        LOG.info('Bazis BG Scheduler started')

        healthcheck_server(
            settings.BAZIS_TASK_SCHEDULER_HOST or socket.gethostname(),
            settings.BAZIS_TASK_SCHEDULER_PORT,
        )
        LOG.info(
            f'Healthcheck server started at {settings.BAZIS_TASK_SCHEDULER_HOST or socket.gethostname()}:{settings.BAZIS_TASK_SCHEDULER_PORT}'
        )

        self.start_handlers_check()

        LOG.info('Entering main loop')

        while True:
            self.run_handlers()

            # perform cron tasks check
            self.cron_check()

            # perform tasks check
            self.tasks_check()

            # perform process start
            self.tasks_prepare()

            # delay
            sleep(settings.BAZIS_TASK_SCHEDULER_IDLE)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile

from django.core.management import call_command
from django.db import transaction

from bazis.contrib.bg.basic.base import BgBase


class FixtureLoad(BgBase):
    name = 'Fixture loading'
    parallel = 1
    clear_models = []

    def handle(self):
        with transaction.atomic():
            # clear models
            for model in self.clear_models:
                model.objects.all().delete()

            # get file
            fp = self.task_file_get()

            # create a temporary local file, since the file may be taken from S3
            tmp_fp = tempfile.NamedTemporaryFile(suffix='.json')
            tmp_fp.write(fp.read())
            tmp_fp.seek(0)

            try:
                call_command('loaddata', tmp_fp.name, ignorenonexistent=True)
            finally:
                tmp_fp.close()
